////////////////////////////////////////////////////////////////////////////////
//      Copyright (c) 2020 Nokia. All Rights Reserved.
//      THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF Nokia.
//      The copyright notice above does not evidence any actual or
//      intended publication of such source code.
//
//      Description: FPC401 driver source file
//
////////////////////////////////////////////////////////////////////////////////

#include "util/Vlog.h"
#include <errno.h>

#include "CFpc40x16p200.h"
#include "DeviceNames.h"
#include "DriverApi.h"
#include "ot16p200_gpio.h"
#include "CDevLayer.h"
#include "DevNames16p200.h"


namespace ot16P200
{
bool CFpc40x16p200::debug_flag = false;
// SFP+ 3~14
#define ASSERT_SFP_PORT(port)                                                                               \
    do                                                                                                      \
    {                                                                                                       \
        if ((port < 3) || (port > 14))                                                                      \
        {                                                                                                   \
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: ERROR! Invalid sfp+ port=%d!\n", __PRETTY_FUNCTION__, port);   \
            return (RV_FAILURE);                                                                            \
        }                                                                                                   \
    } while (0)

// QSFP physical port 1,2,15,16,17,18,19,20,21,22
#define ASSERT_QSFP_PHY_PORT(port)                                                                                     \
    do                                                                                                                 \
    {                                                                                                                  \
        if (!((port == 1) || (port == 2) || ((port >= 15) && (port <= 22))))                                           \
        {                                                                                                              \
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: ERROR! Invalid qsfp28 physical port=%d!\n", __PRETTY_FUNCTION__, port);   \
            return (RV_FAILURE);                                                                                       \
        }                                                                                                              \
    } while (0)

#ifndef SIMULATION
// The FPC401 devices are connected in a daisy-chain fashion.
// The first FPC401 will connect CTRL3 (MOSI) to the MOSI signal of the host controller.
// Based on the Figure 13, CPOL=0 CPHA=0, so the SPI mode is 0
CFpc40x16p200::CFpc40x16p200(const char *pName, int busName, const char *devName) throw(DLRC)
    : ubcmDevice::CspiAdapter(pName, busName, devName, 0, 8)
{
    // Initialization Variables for SPI
    // inst 0, 1, 2 for 12 SFPs, inst 3, 4 for client QSFP, inst 5 for line QSFP.
    uint8_t inst_addr_map[6] = { 0, 1, 2, 3, 4, 5 };
    uint8_t port_fpc_dev_addr_map[24] = { 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3,
                                          4, 4, 4, 4, 5, 5, 5, 5 };
    uint8_t port_fpc_local_port_map[24] = { 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3,
                                            0, 1, 2, 3, 0, 1, 2, 3 };

    // Create FPC structure
    // SPI comm_type(1), 6 FPC401 devices, 24 port numbers.
    pFpcDevice = fpc401_device_start(&CFpc40x16p200::spi_write, &CFpc40x16p200::spi_read, 1, 6, 24,
            inst_addr_map, port_fpc_dev_addr_map, port_fpc_local_port_map);
    if (pFpcDevice == NULL)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "CFpc40x16p200::CFpc40x16p200: can't create FPC structure.\n");
        throw (DLRC_ERROR);
    }
    else
    {
        LTRACE(TRCOM|TRCHIGH, "CFpc40x16p200::CFpc40x16p200: create FPC structure success.\n");
    }
}
#else
CFpc40x16p200::CFpc40x16p200(const char *pName, int busName, const char *devName) throw(DLRC)
    : ubcmDevice::CDevice(pName, -1, NULL)
{
    LTRACE(TRCOM_ERROR|TRCHIGH, "CFpc40x16p200::CFpc40x16p200: pname %s busName %d devName %s", pName, busName, devName);

    // Initialization Variables for SPI
    // inst 0, 1, 2 for 12 SFPs, inst 3, 4 for client QSFP, inst 5 for line QSFP.
    uint8_t inst_addr_map[6] = { 0, 1, 2, 3, 4, 5 };
    uint8_t port_fpc_dev_addr_map[24] = { 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3,
                                          4, 4, 4, 4, 5, 5, 5, 5 };
    uint8_t port_fpc_local_port_map[24] = { 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3,
                                            0, 1, 2, 3, 0, 1, 2, 3 };

    // Create FPC structure
    // SPI comm_type(1), 6 FPC401 devices, 24 port numbers.
    pFpcDevice = fpc401_device_start(&CFpc40x16p200::spi_write, &CFpc40x16p200::spi_read, 1, 6, 24,
            inst_addr_map, port_fpc_dev_addr_map, port_fpc_local_port_map);
    if (pFpcDevice == NULL)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "CFpc40x16p200::CFpc40x16p200: can't create FPC structure.\n");
        throw (DLRC_ERROR);
    }
    else
    {
        LTRACE(TRCOM|TRCHIGH, "CFpc40x16p200::CFpc40x16p200: create FPC structure success.\n");
    }
}
#endif

CFpc40x16p200::~CFpc40x16p200()
{
    fpc401_device_stop(*pFpcDevice);
}

RetVal CFpc40x16p200::init() // called only when cold reboot
{
    const char *fn = "CFpc40x16p200::init";
    for (uint8_t port = 0; port < OPTICAL_MODULES_LED_NUM; port++)
    {
        // configure green led state to off when init
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_GREEN_LED, FPC_LED_OFF, 1, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: fail to configure port %d LED %d off\n", fn, port + 1, FPC401_SFP_GREEN_LED);
            return RV_FAILURE;
        }
        else
        {
            unreserveSpiBus();
            LTRACE(TRCOM|TRCLOW, "%s: configure port %d LED %d off success\n", fn, port + 1, FPC401_SFP_GREEN_LED);
        }

        // configure green led brightness to 100% when init
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_brightness_set(*pFpcDevice, port, FPC401_SFP_GREEN_LED, 100, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: fail to configure port %d LED %d brightness\n", fn, port + 1, FPC401_SFP_GREEN_LED);
            return RV_FAILURE;
        }
        else
        {
            unreserveSpiBus();
            LTRACE(TRCOM|TRCLOW, "%s: configure port %d LED %d brightness success\n", fn, port + 1, FPC401_SFP_GREEN_LED);
        }

        // configure green led blink on time of 500 ms, off time of 500 ms
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_blink_set(*pFpcDevice, port, FPC401_SFP_GREEN_LED, 200, 200, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: fail to configure port %d LED %d blink period\n", fn, port + 1, FPC401_SFP_GREEN_LED);
            return RV_FAILURE;
        }
        else
        {
            unreserveSpiBus();
            LTRACE(TRCOM|TRCLOW, "%s: configure port %d LED %d blink period success\n", fn, port + 1, FPC401_SFP_GREEN_LED);
        }

        // configure red led state to off when init
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_RED_LED, FPC_LED_OFF, 1, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: fail to configure port %d LED %d off\n", fn, port + 1, FPC401_SFP_RED_LED);
            return RV_FAILURE;
        }
        else
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: configure port %d LED %d off success\n", fn, port + 1, FPC401_SFP_RED_LED);
        }

        // configure red led brightness to 100% when init
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_brightness_set(*pFpcDevice, port, FPC401_SFP_RED_LED, 100, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: fail to configure port %d LED %d brightness\n", fn, port + 1, FPC401_SFP_RED_LED);
            return RV_FAILURE;
        }
        else
        {
            unreserveSpiBus();
            LTRACE(TRCOM|TRCLOW, "%s: configure port %d LED %d brightness success\n", fn, port + 1, FPC401_SFP_RED_LED);
        }

        // configure red led blink on time of 500 ms, off time of 500 ms
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_blink_set(*pFpcDevice, port, FPC401_SFP_RED_LED, 200, 200, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: fail to configure port %d LED %d blink period\n", fn, port, FPC401_SFP_RED_LED);
            return RV_FAILURE;
        }
        else
        {
            unreserveSpiBus();
            LTRACE(TRCOM|TRCLOW, "%s: configure port %d LED %d blink period success\n", fn, port + 1, FPC401_SFP_RED_LED);
        }

        // configure LED clock forwarding
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_clock_forward(*pFpcDevice))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: fail to configure port %d LED clock forwarding\n", fn, port + 1);
            return RV_FAILURE;
        }
        else
        {
            unreserveSpiBus();
            LTRACE(TRCOM|TRCLOW, "%s: configure port %d LED clock forwarding success\n", fn, port + 1);
        }

        // reset LED counters
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_sync(*pFpcDevice, (FPC401_INSTANCE)(port/4), 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: fail to configure port %d LED sync\n", fn, port + 1);
            return RV_FAILURE;
        }
        else
        {
            unreserveSpiBus();
            LTRACE(TRCOM|TRCLOW, "%s: configure port %d LED sync success\n", fn, port + 1);
        }
    }

    // QSFP28 Client 1, QSFP28 Client 2, QSFP28 L1 P0 i2c watchdog timeout register, set the timeout to 6ms
    for (uint16_t devId = FPC_INSTANCE_3; devId < FPC_INSTANCE_MAX; devId++)
    {
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->write_fn(devId, FPC401_P0_MASTER_WATCHDOG, 0x6))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s device %d failed to write port I2C timeout register 0xA9!!!\n", fn, devId);
            return (RV_FAILURE);
        }
        unreserveSpiBus();
    }

    // QSFP L2 P1 i2c wachdog timeout register, set the timeout to 6ms
    reserveSpiBus();
    if (RV_SUCCESS != pFpcDevice->write_fn(FPC_INSTANCE_5, FPC401_P1_MASTER_WATCHDOG, 0x6))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s device %d failed to write port I2C timeout register 0xAA!!!\n", fn, FPC_INSTANCE_5);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    return RV_SUCCESS;
}

void CFpc40x16p200::constructSpiReadFrame(uint16_t inst, uint16_t reg, uint8_t *frame)
{
    // 6*29 bits = 174 bits, 21 bytes 6 bits(+2 additional bits)-> 22 bytes
    // 2 additional bits + 29 bits for 6th device cascaded + 29 bits for 5th device cascaded + 
    // 29 bits for 4th device cascaded + 29 bits for 3rd device cascaded +
    // 29 bits for 2nd device cascaed + 29 bits for 1st device cascaed
    // Note: Bits get shifted out from master SPI device from left to right

    // byte 1: two junk bits 0 + read command bit<1>(6th device) + 5 bits addr(6th device)
    // byte 2: 7 bits addr(6th device) + 1 bit dummy data(6th device)
    // byte 3: 8 bits dummy data(6th device)
    // byte 4: 7 bits dummy data(6th device) + read command bit<1>(5th device)
    // byte 5: 8 bits addr(5th device)
    // byte 6: 4 bits addr(5th device) + 4 bits dummy data(5th device)
    // byte 7: 8 bits dummy data(5th device)
    // byte 8: 4 bits dummy data(5th device) + read command bit<1>(4th device) + 3 bits addr(4th device)
    // byte 9: 8 bits addr(4th device)
    // byte 10: 1 bit addr(4th device) + 7 bits dummy data(4th device)
    // byte 11: 8 bits dummy data(4th device)
    // byte 12: 1 bit dummy data(4th device) + read command bit<1>(3rd device) + 6 bits addr(3rd device)
    // byte 13: 6 bits addr(3rd device) + 2 bits dummy data(3rd device)
    // byte 14: 8 bits dummy data(3rd device)
    // byte 15: 6 bits dummy data(3rd device) + read command bit<1>(2nd device) + 1 bit addr(2nd device)
    // byte 16: 8 bits addr(2nd device)
    // byte 17: 3 bits addr(2nd device) + 5 bit dummy data(2nd device)
    // byte 18: 8 bits dummy data(2nd device)
    // byte 19: 3 bits dummy data(2nd device) + read command bit<1>(1st device) + 4 bits addr(1st device)
    // byte 20: 8 bits addr(1st device)
    // byte 21: 8 bits dummy data(1st device)
    // byte 22: 8 bits dummy data(1st device)
    LTRACE(TRCOM|TRCLOW, "CFpc40x16p200::constructSpiReadFrame: inst %d reg %x\n", inst, reg);

    switch (inst)
    {
        case FPC_INSTANCE_0:
            // 3 bits dummy data + read command bit<1> + 4 bits addr
            frame[18] = 0xf0 | (uint8_t)((reg & 0x0f00) >> 8);           // for the FPC401 1 cascaded
            // 8 bits addr
            frame[19] = (uint8_t)(reg & 0x00ff);
            // 8 bits dummy data
            frame[20] = 0x00;
            // 8 bits dummy data
            frame[21] = 0x00;
            break;

        case FPC_INSTANCE_1:
            // 6 bits dummy data + read command bit<1> + 1 bit addr
            frame[14] = 0xfe | (uint8_t)((reg & 0x0800) >> 11);           // for the FPC401 2 cascaded
            // 8 bits addr
            frame[15] = (uint8_t)((reg & 0x07f8) >> 3);
            // 3 bits addr + 5 bit dummy data
            frame[16] = (uint8_t)(((reg & 0x0007) << 5) & 0xe0);
            // 8 bits dummy data
            frame[17] = 0x00;
            // 3 bits dummy data
            frame[18] = 0x1f;
            break;

        case FPC_INSTANCE_2:
            // 1 bit dummy data + read command bit<1> + 6 bits addr
            frame[11] = 0xc0 | (uint8_t)((reg & 0x0fc0) >> 6);             // for the FPC401 3 cascaded
            // 6 bits addr + 2 bits dummy data
            frame[12] = (uint8_t)(((reg & 0x003f) << 2) & 0xfc);
            // 8 bits dummy data
            frame[13] = 0x00;
            // 6 bits dummy data
            frame[14] = 0x03;
            break;

        case FPC_INSTANCE_3:
            // 4 bits dummy data + read command bit<1> + 3 bits addr
            frame[7] = 0xf8 | (uint8_t)((reg & 0x0e00) >> 9);               // for the FPC401 4 cascaded
            // 8 bits addr
            frame[8] = (uint8_t)((reg & 0x01fe) >> 1);
            // 1 bit addr + 7 bits dummy data
            frame[9] = (uint8_t)(((reg & 0x0001) << 7) & 0x0080);
            // 8 bits dummy data
            frame[10] = 0x00;
            // 1 bit dummy data
            frame[11] = 0x7f;
            break;

        case FPC_INSTANCE_4:
            // 7 bits dummy data + read command bit<1>
            frame[3] = 0xff;                                                // for the FPC401 5 cascaded
            // 8 bits addr
            frame[4] = (uint8_t)((reg & 0x0ff0) >> 4);
            // 4 bits addr + 4 bits dummy data
            frame[5] = (uint8_t)(((reg & 0x000f) << 4) & 0x00f0);
            // 8 bits dummy data
            frame[6] = 0x00;
            // 4 bits dummy data
            frame[7] = 0x0f;
            break;

        case FPC_INSTANCE_5:
            // two junk bits 0 + read command bit<1> + 5 bits addr           // 2 addtional bits
            frame[0] = 0x20 | (uint8_t)((reg & 0x0f80) >> 7);                // for the FPC401 6 cascaded
            // 7 bits addr + 1 bit dummy data
            frame[1] = (uint8_t)(((reg & 0x007f) << 1) & 0x00fe);
            // 8 bits dummy data
            frame[2] = 0x00;
            // 7 bits dummy data
            frame[3] = 0x01;
            break;

        default:
            LTRACE(TRCOM_ERROR|TRCHIGH, "The FPC instance number %d is not correct", inst);
            break;
    }
}

void CFpc40x16p200::constructSpiWriteFrame(uint16_t inst, uint16_t reg, uint8_t data, uint8_t *frame)
{
    // 6*29 bits = 174 bits, 21 bytes 6 bits(+2 additional bits)-> 22 bytes
    // 2 additional bits + 29 bits for 6th device cascaded + 29 bits for 5th device cascaded + 
    // 29 bits for 4th device cascaded + 29 bits for 3rd device cascaded +
    // 29 bits for 2nd device cascaed + 29 bits for 1st device cascaed
    // Note: Bits get shifted out from master SPI device from left to right

    // byte 1: two junk bits 0 + write command bit<0>(6th device) + 5 bits addr(6th device)
    // byte 2: 7 bits addr(6th device) + 1 bit data(6th device)
    // byte 3: 8 bits data(6th device)
    // byte 4: 7 bits data(6th device) + write command bit<0>(5th device)
    // byte 5: 8 bits addr(5th device)
    // byte 6: 4 bits addr(5th device) + 4 bits data(5th device)
    // byte 7: 8 bits data(5th device)
    // byte 8: 4 bits data(5th device) + write command bit<0>(4th device) + 3 bits addr(4th device)
    // byte 9: 8 bits addr(4th device)
    // byte 10: 1 bit addr(4th device) + 7 bits data(4th device)
    // byte 11: 8 bits data(4th device)
    // byte 12: 1 bit data(4th device) + write command bit<0>(3rd device) + 6 bits addr(3rd device)
    // byte 13: 6 bits addr(3rd device) + 2 bits data(3rd device)
    // byte 14: 8 bits data(3rd device)
    // byte 15: 6 bits data(3rd device) + write command bit<0>(2nd device) + 1 bit addr(2nd device)
    // byte 16: 8 bits addr(2nd device)
    // byte 17: 3 bits addr(2nd device) + 5 bits data(2nd device)
    // byte 18: 8 bits data(2nd device)
    // byte 19: 3 bits data(2nd device) + write command bit<0>(1st device) + 4 bits addr(1st device)
    // byte 20: 8 bits addr(1st device)
    // byte 21: 8 bits data(1st device)
    // byte 22: 8 bits data(1st device)
    LTRACE(TRCOM|TRCLOW, "CFpc40x16p200::constructSpiWriteFrame: inst %d reg %x data %x\n", inst, reg, data);
    switch (inst)
    {
        case FPC_INSTANCE_0:
            // 3 bits dummy data + write command bit<0> + 4 bits addr
            frame[18] = 0xe0 | (uint8_t)((reg & 0x0f00) >> 8);           // for the FPC401 1 cascaded
            // 8 bits addr
            frame[19] = (uint8_t)(reg & 0x00ff);
            // 8 bits dummy data
            frame[20] = 0x00;
            // 8 bits sending data
            frame[21] = data;
            break;

        case FPC_INSTANCE_1:
            // 6 bits dummy data + write command bit<0> + 1 bit addr
            frame[14] = 0xfc | (uint8_t)((reg & 0x0800) >> 11);           // for the FPC401 2 cascaded
            // 8 bits addr
            frame[15] = (uint8_t)((reg & 0x07f8) >> 3);
            // 3 bits addr + 5 bits dummy data
            frame[16] = (uint8_t)(((reg & 0x0007) << 5) & 0x00e0);
            // 3 bits dummy data + 5 bits sending data
            frame[17] = (uint8_t)((data & 0xf8) >> 3) & 0x1f;
            // 3 bits sending data + 5 bits dummy data
            frame[18] = (uint8_t)((data & 0x07) << 5) | 0x1f;
            break;

        case FPC_INSTANCE_2:
            // 1 bit dummy data + write command bit<0> + 6 bits addr
            frame[11] = 0x80 | (uint8_t)((reg & 0x0fc0) >> 6);             // for the FPC401 3 cascaded
            // 6 bits addr + 2 bits dummy data
            frame[12] = (uint8_t)(((reg & 0x003f) << 2) & 0x00fc);
            // 6 bits dummy data + 2 bits sending data
            frame[13] = (uint8_t)((data & 0xc0) >> 6) & 0x03;
            // 6 bits sending data + 2 bits dummy data
            frame[14] = (uint8_t)((data & 0x3f) << 2) | 0x03;
            break;

        case FPC_INSTANCE_3:
            // 4 bits dummy data + write command bit<0> + 3 bits addr
            frame[7] = 0xf0 | (uint8_t)((reg & 0x0e00) >> 9);               // for the FPC401 4 cascaded
            // 8 bits addr
            frame[8] = (uint8_t)((reg & 0x01fe) >> 1);
            // 1 bit addr + 7 bits dummy data
            frame[9] = (uint8_t)(((reg & 0x0001) << 7) & 0x0080);
            // 1 bit dummy data + 7 bits sending data
            frame[10] = (uint8_t)((data & 0xfe) >> 1) & 0x7f;
            // 1 bit sending data + 7 bits dummy data
            frame[11] = (uint8_t)((data & 0x01) << 7) | 0x7f;
            break;

        case FPC_INSTANCE_4:
            // 7 bits dummy data + write command bit<0>                // for the FPC401 5 cascaded
            frame[3] = 0xfe;
            // 8 bits addr
            frame[4] = (uint8_t)((reg & 0x0ff0) >> 4);
            // 4 bits addr + 4 bits dummy data
            frame[5] = (uint8_t)(((reg & 0x000f) << 4) & 0x00f0);
            // 4 bits dummy data + 4 bits sending data
            frame[6] = (uint8_t)((data & 0xf0) >> 4) & 0x0f;
            // 4 bits sending data + 4 bits dummy data
            frame[7] = (uint8_t)((data & 0x0f) << 4) | 0x0f;
            break;

        case FPC_INSTANCE_5:
            // two junk bits 0 + write command bit<0> + 5 bits addr    // 2 addtional bits
            frame[0] = 0x00 | (uint8_t)((reg & 0x0f80) >> 7);          // for the FPC401 6 cascaded
            // 7 bits addr + 1 bit dummy data
            frame[1] = (uint8_t)(((reg & 0x007f) << 1) & 0xfe);
            // 7 bits dummy data + 1 bit sending data
            frame[2] = 0x00 | (uint8_t)((data & 0x80) >> 7);
            // 7 bits sending data + 1 bit dummy data
            frame[3] = (uint8_t)((data & 0x7f) << 1) | 0x01;
            break;

        default:
            LTRACE(TRCOM_ERROR|TRCHIGH, "The FPC instance number %d is not correct", inst);
            break;
    }
}

// Each FPC401 device in the SPI chain will capture and act upon the command in its shift register
// when SS_N transitions from low (0) to high (1).
// The MOSI input is ignored and the MISO output is high impedance whenever SS_N is deasserted high.
// The SPI address space for FPC401 applications is designed such that each port being controlled and each
// logical device address within each port is accessible to the host controller through a unique 12-bit address.
// Each SPI transaction to a single FPC401 device is 29 bits long and is framed by the assertion of SS_N (CTRL2) low.
// BIT    FIELD      DESCRIPTION
// 28     R/W        0: Write command 1: Read command. This is the first bit shifted in on the MOSI input.
// 27:16  ADDR[11:0] 12-bit address field.
// 15     DATA[15]   Busy flag. For read operations, a 1 means the downstream port is busy. For write, don't care.
// 14     DATA[14]   Don't care.
// 13     DATA[13]   NACK received flag. A 1 means the FPC401 has received a NACK from the downstream port.
// 12     DATA[12]   Reject flag. A 1 means the previous command rejected because FPC401 is busy servicing a prior command.
// 11:8   DATA[11:8] Don't care.
// 7:0    DATA[7:0]  8-bit data field. DATA[0] is the last bit shifted in on the MOSI input.
// Inside FPC SPI interface there is a 29-bits shift register, use 32 bits or multiple of 8 bits. 
// When sending to the FPC device, the first 3 bits could be junk or don’t care.
// because these would get shifted out of the shift register.
// On the other hand, on the receive side or when reading from FPC, the last three bits would be don’t care.
// Essentially every 29-bits frame belongs to one instance of FPC device.
// If there are more FPC devices then each one would have additional 29-bits frame.
cerrno CFpc40x16p200::spi_read(uint16_t inst, uint16_t reg, uint8_t *data)
{
    const char *fn = "CFpc40x16p200::spi_read";
    LTRACE(TRCOM|TRCLOW, "%s: inst %x reg %x\n", fn, inst, reg);

    if ((inst != FPC_INSTANCE_0) && (inst != FPC_INSTANCE_1) && (inst != FPC_INSTANCE_2) &&
        (inst != FPC_INSTANCE_3) && (inst != FPC_INSTANCE_4) && (inst != FPC_INSTANCE_5))
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: invalid inst %x.\n", fn, inst);
        return (EINVAL);
    }

    CFpc40x16p200 *pFpc401 = static_cast<CFpc40x16p200 *>(CDevLayer::findDevice(NAME_OPTICS_CTRL_FPC401));
    if (pFpc401 == NULL)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: Can not find the FPC401 Device Driver!\n", fn);
        return (ENXIO);
    }

    // Reading data from an FPC401 device requires two complete SPI transactions

#ifndef SIMULATION
    int retries1 = 0;
    int retries2 = 0;

    // the first SPI transaction
    // construct the SPI frame structure
    uint8_t cmd[FPC401_SPI_FRAME_BYTES];
    memset(cmd, 0xff, FPC401_SPI_FRAME_BYTES);
    // The first two junk bits need to be 0
    cmd[0] = 0x3f;
    constructSpiReadFrame(inst, reg, cmd);

    if (debug_flag)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "cmd %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x",
        cmd[0], cmd[1], cmd[2], cmd[3], cmd[4], cmd[5], cmd[6], cmd[7], cmd[8], cmd[9], cmd[10], cmd[11],
        cmd[12], cmd[13], cmd[14], cmd[15], cmd[16], cmd[17], cmd[18], cmd[19], cmd[20], cmd[21]);
    }

    nanosleep((const struct timespec[]){ { 0, 5000000 } }, NULL); // wait 5ms

    SetGpio(GPIO_SPLD_SS_N);
    SetGpio(GPIO_FPC401_SS_N);
    nanosleep((const struct timespec[]){ { 0, 10000} }, NULL); // wait 10us
    ClearGpio(GPIO_FPC401_SS_N);

    for (retries1 = 0; retries1 < FPC401_SPI_RETRIES; retries1++)
    {
        if (pFpc401->SpiXfer((char *)cmd, FPC401_SPI_FRAME_BYTES) == 0) // cmd size is 22 bytes
        {
            break;
        }
        else
        {
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: SpiXfer failed.\n", fn);
        }
    }

    // inactive and reactive CS
    SetGpio(GPIO_FPC401_SS_N);
    // the required time delay between the two transactions
    if (reg >= 0x800) // Local FPC402 register reads: tOFF-SSN ≥ 1 μs
    {
        nanosleep((const struct timespec[]){ { 0, 10000 } }, NULL); // wait 10us
    }
    else // Downstream (remote) register reads: tOFF-SSN ≥ 170 μs assuming 400-kHz I2C; 620 μs assuming 100-kHz I2C
    {
        if ((inst == FPC_INSTANCE_0) || (inst == FPC_INSTANCE_1) || (inst == FPC_INSTANCE_2))
        {
            nanosleep((const struct timespec[]){ { 0, 1000000 } }, NULL); // wait 1ms
        }
        else
        {
            nanosleep((const struct timespec[]){ { 0, 1200000 } }, NULL); // wait 1.2ms
        }
    }
    ClearGpio(GPIO_FPC401_SS_N);

    // the second SPI transaction
    // A dummy frame composed of all ones.
    // The first two junk bits needs to be 0.
    cmd[0] = 0x3f;
    for (int i = 1; i < FPC401_SPI_FRAME_BYTES; i++)
    {
        cmd[i] = 0xff;
    }

    for (retries2 = 0; retries2 < FPC401_SPI_RETRIES; retries2++)
    {
        if (pFpc401->SpiXfer((char *)cmd, FPC401_SPI_FRAME_BYTES) == 0) // cmd size is 22 bytes
        {
            break;
        }
        else
        {
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: SpiXfer failed.\n", fn);
        }
    }
    SetGpio(GPIO_FPC401_SS_N);

    if (debug_flag)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "cmd %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x",
        cmd[0], cmd[1], cmd[2], cmd[3], cmd[4], cmd[5], cmd[6], cmd[7], cmd[8], cmd[9], cmd[10], cmd[11],
        cmd[12], cmd[13], cmd[14], cmd[15], cmd[16], cmd[17], cmd[18], cmd[19], cmd[20], cmd[21]);
    }

    if ((retries1 == FPC401_SPI_RETRIES) || (retries2 == FPC401_SPI_RETRIES))
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: spi transfer failed after %d %d retries\n", fn, retries1, retries2);
        return (EIO);
    }
    else
    {
        if ((retries1 > 0) || (retries2 > 0))
        {
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: recovered after %d %d retries\n", fn, retries1, retries2);
        }
    }

    // 6 SPI cascaded device response
    // 29 bits from device 6 + 29 bits from device 5 + 29 bits from device 4 +
    // 29 bits from device 3 + 29 bits from device 2 + 29 bits from device 1 + 2 bits don't care
    // byte 1: read command bit<1>(6th device) + 7 bits addr(6th device)
    // byte 2: 5 bits addr(6th device) + 3 bit data(6th device)
    // byte 3: 8 bits data(6th device)
    // byte 4: 5 bits data(6th device) + read command bit<1>(5th device) + 2 bit addr(5th device)
    // byte 5: 8 bits addr(5th device)
    // byte 6: 2 bits addr(5th device) + 6 bits data(5th device)
    // byte 7: 8 bits data(5th device)
    // byte 8: 2 bits data(5th device) + read command bit<1>(4th device) + 5 bits addr(4th device)
    // byte 9: 7 bits addr(4th device) + 1 bit data(4th device)
    // byte 10: 8 bits data(4th device)
    // byte 11: 7 bits data(4th device) + read command bit<1>(3rd device)
    // byte 12: 8 bits addr(3rd device)
    // byte 13: 4 bits addr(3rd device) + 4 bits data(3rd device)
    // byte 14: 8 bits data(3rd device)
    // byte 15: 4 bits data(3rd device) + read command bit<1>(2nd device) + 3 bit addr(2nd device)
    // byte 16: 8 bits addr(2nd device)
    // byte 17: 1 bits addr(2nd device) + 7 bits data(2nd device)
    // byte 18: 8 bits data(2nd device)
    // byte 19: 1 bit data(2nd device) + read command bit<1>(1st device) + 6 bits addr(1st device)
    // byte 20: 6 bits addr(1st device) + 2 bits data(1st device)
    // byte 21: 8 bits data(1st device)
    // byte 22: 6 bits data(1st device) + 2 bits junk bits
    uint8_t status = 0;
    switch (inst)
    {
        case FPC_INSTANCE_0:
            *data = (uint8_t)(((cmd[20] & 0x03) << 6) | ((cmd[21] & 0xfc) >> 2));
            status = (uint8_t)(((cmd[19] & 0x03) << 6) | ((cmd[20] & 0xfc) >> 2));
            break;
        case FPC_INSTANCE_1:
            *data = (uint8_t)(((cmd[17] & 0x7f) << 1) | ((cmd[18] & 0x80) >> 7));
            status = (uint8_t)(((cmd[16] & 0x7f) << 1) | ((cmd[17] & 0x80) >> 7));
            break;
        case FPC_INSTANCE_2:
            *data = (uint8_t)(((cmd[13] & 0x0f) << 4) | ((cmd[14] & 0xf0) >> 4));
            status = (uint8_t)(((cmd[12] & 0x0f) << 4) | ((cmd[13] & 0xf0) >> 4));
            break;
        case FPC_INSTANCE_3:
            *data = (uint8_t)(((cmd[9] & 0x01) << 7) | ((cmd[10] & 0xfe) >> 1));
            status = (uint8_t)(((cmd[8] & 0x01) << 7) | ((cmd[9] & 0xfe) >> 1));
            break;
        case FPC_INSTANCE_4:
            *data = (uint8_t)(((cmd[6] & 0x3f) << 2) | ((cmd[7] & 0xc0) >> 6));
            status = (uint8_t)(((cmd[5] & 0x3f) << 2) | ((cmd[6] & 0xc0) >> 6));
            break;
        case FPC_INSTANCE_5:
            *data = (uint8_t)(((cmd[2] & 0x07) << 5) | ((cmd[3] & 0xf8) >> 3));
            status = (uint8_t)(((cmd[1] & 0x07) << 5) | ((cmd[2] & 0xf8) >> 3));
            break;
    }

    LTRACE(TRCOM|TRCLOW, "%s: data %x, check status %x\n", fn, *data, status);

    // Data[15], Busy flag. For read operations, a 1 means the downstream port is busy.
    if ((status & 0x80) != 0)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: the downstream port is busy.\n", fn);
        return (EBUSY);
    }

    // Data[13], NACK received flag. A 1 means the FPC402 has received a NACK from the downstream port.
    if ((status & 0x20) != 0)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: received a NACK from the downstream port.\n", fn);
        return (EAGAIN);
    }

    // DATA[12], Reject flag. A 1 means the FPC402 has rejected the previous command because it is busy servicing a prior command.
    if ((status & 0x10) != 0)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: rejected the previous command.\n", fn);
        return (EBUSY);
    }
#else
    *data = 0;
#endif

    return (EOK);
}

cerrno CFpc40x16p200::spi_write(uint16_t inst, uint16_t reg, uint8_t data)
{
    const char *fn = "CFpc40x16p200::spi_write";
    LTRACE(TRCOM|TRCLOW, "%s: inst %x reg %x data %x\n", fn, inst, reg, data);

    if ((inst != FPC_INSTANCE_0) && (inst != FPC_INSTANCE_1) && (inst != FPC_INSTANCE_2) &&
        (inst != FPC_INSTANCE_3) && (inst != FPC_INSTANCE_4) && (inst != FPC_INSTANCE_5))
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: invalid addr %d.\n", fn, inst);
        return (EINVAL);
    }

    CFpc40x16p200 *pFpc401 = static_cast<CFpc40x16p200 *>(CDevLayer::findDevice(NAME_OPTICS_CTRL_FPC401));
    if (pFpc401 == NULL)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: Can not find the FPC401 Device Driver!\n", fn);
        return (ENODEV);
    }

#ifndef SIMULATION
    int retries = 0;

    // Writing data to an FPC401 device or the downstream ports requires one SPI transactions.
    // construct the SPI frame structure
    // 6*29 bits = 174 bits, 21 bytes 6 bits -> 22 bytes
    uint8_t cmd[22];
    memset(cmd, 0xff, FPC401_SPI_FRAME_BYTES);
    cmd[0] = 0x3f;
    constructSpiWriteFrame(inst, reg, data, cmd);
    if (debug_flag)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "cmd %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x",
        cmd[0], cmd[1], cmd[2], cmd[3], cmd[4], cmd[5], cmd[6], cmd[7], cmd[8], cmd[9], cmd[10], cmd[11],
        cmd[12], cmd[13], cmd[14], cmd[15], cmd[16], cmd[17], cmd[18], cmd[19], cmd[20], cmd[21]);
    }

    nanosleep((const struct timespec[]){ { 0, 5000000 } }, NULL); // wait 5ms

    SetGpio(GPIO_SPLD_SS_N);
    SetGpio(GPIO_FPC401_SS_N);
    nanosleep((const struct timespec[]){ { 0, 10000} }, NULL); // wait 10us
    ClearGpio(GPIO_FPC401_SS_N);

    for (retries = 0; retries < FPC401_SPI_RETRIES; retries++)
    {
        if (pFpc401->SpiXfer((char *)cmd, FPC401_SPI_FRAME_BYTES) == 0) // cmd size is 22 bytes
        {
            break;
        }
        else
        {
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: SpiXfer failed.\n", fn);
        }
    }
    
    SetGpio(GPIO_FPC401_SS_N);

    if (retries == FPC401_SPI_RETRIES)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: spi transfer failed after %d retries\n", fn, retries);
        return (ENXIO);
    }
    else
    {
        if (retries > 0)
        {
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: SpiXfer recovered after %d retries\n", fn, retries);
        }
    }
#endif

    return (EOK);
}

// Reserved register write and read back for sanity check
RetVal CFpc40x16p200::sanityTest()
{
    const char *fn = "CFpc40x16p200::sanityTest:";
    uint8_t testData = 0x55;
    uint8_t valueBack = 0x0;

    for (uint16_t devId = 0; devId < FPC_INSTANCE_MAX; devId++)
    {
        reserveSpiBus();
        // Reserved Register (offset = 0xB1)
        if (RV_SUCCESS != pFpcDevice->write_fn(devId, FPC401_RESERVED_REG, testData))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s device %d failed to write reserved register 0xB1!!!\n", fn, devId);
            return (RV_FAILURE);
        }
        unreserveSpiBus();

        nanosleep((const struct timespec[]){ { 0, 10000000L } }, NULL); // wait 10ms

        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->read_fn(devId, FPC401_RESERVED_REG, &valueBack))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s device %d failed to read reserved register 0xB1!!!\n", fn, devId);
            return (RV_FAILURE);
        }
        unreserveSpiBus();

#ifndef SIMULATION
        if (valueBack != testData)
        {
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s device %d sanity test failed, valueBack=%x!!!\n", fn, devId, valueBack);
            return (RV_FAILURE);
        }
        else
        {
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s device %d sanity test success %x = %x!!!\n", fn, devId, testData, valueBack);
        }
#endif
    }

    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::getVersionId()
{
    cerrno ret = RV_SUCCESS;

    uint8_t version = 0;
    for (uint16_t instNum = 0; instNum < FPC_INSTANCE_MAX; instNum++)
    {
        reserveSpiBus();
        ret = pFpcDevice->read_fn(instNum, FPC401_DEVICE_VERSION, &version);
        if (ret != EOK)
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "CFpc40x16p200::getVersionId: read inst %d device version failed\n", instNum);
            return (RV_FAILURE);
        }
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "CFpc40x16p200::getVersionId: inst %d device version %x\n", instNum, version);
    }

    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::getDevIdNum()
{
    cerrno ret = RV_SUCCESS;

    uint8_t devId = 0;
    for (uint16_t instNum = 0; instNum < FPC_INSTANCE_MAX; instNum++)
    {
        reserveSpiBus();
        ret = pFpcDevice->read_fn(instNum, FPC401_DEVICE_ID, &devId);
        if (ret != EOK)
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "CFpc40x16p200::getDevIdNum: read inst %d device id failed\n", instNum);
            return (RV_FAILURE);
        }
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "CFpc40x16p200::getDevIdNum: inst %d device id %x\n", instNum, devId);
    }    

    return (RV_SUCCESS);
}

void CFpc40x16p200::convertQsfpPort(int qsfpPort, uint8_t &port, bool isLedCfg)
{
    // Line Port, L1 1, L2 2
    if (qsfpPort == 1 || qsfpPort == 2)
    {
        port = (uint8_t)(qsfpPort + 19);
    }
    // Client Port, C1 15, C2 19
    else if (qsfpPort == 15 || qsfpPort == 16 || qsfpPort == 17 || qsfpPort == 18)
    {
        if (isLedCfg)
        {
            port = (uint8_t)(qsfpPort - 3);
        }
        else
        {
            port = 12;
        }
    }
    else if (qsfpPort == 19 || qsfpPort == 20 || qsfpPort == 21 || qsfpPort == 22)
    {
        if (isLedCfg)
        {
            port = (uint8_t)(qsfpPort - 3);
        }
        else
        {
            port = 16;
        }
    }
}

// The ModPrsL is asserted "Low" when inserted and deasserted "High" when the module is
// physically absent from the host connector.
RetVal CFpc40x16p200::getQsfpStatPrest(int qsfpPort, bool& qsfpPrest)
{
    const char *fn = "CFpc40x16p200::getQsfpStatPrest:";
    ASSERT_QSFP_PHY_PORT(qsfpPort);
    qsfpPrest = false;

    uint8_t data = 0;
    uint8_t port = 0;
    convertQsfpPort(qsfpPort, port);

    reserveSpiBus();
    if (RV_SUCCESS != pFpcDevice->in_get(*pFpcDevice, port, IN_B_MOD_ABS, &data))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s failed to get qsfpPort=%d absent status\n", fn, qsfpPort);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    if (data != 0)
    {
        qsfpPrest = false;
    }
    else
    {
        qsfpPrest = true;
    }

    return (RV_SUCCESS);
}

// put modules into a low power mode when LPMode pin is high
RetVal CFpc40x16p200::setQsfpCtrlLowPwr(int qsfpPort, bool lowPwr)
{
    const char *fn = "CFpc40x16p200::setQsfpCtrlLowPwr:";
    ASSERT_QSFP_PHY_PORT(qsfpPort);
    LTRACE(TRCOM|TRCLOW, "%s qsfpPort %d, lowPwr %d\n", fn, qsfpPort, lowPwr);

    uint8_t port = 0;
    convertQsfpPort(qsfpPort, port);

    reserveSpiBus();
    // The output values must be configured before the outputs are enabled.
    if (RV_SUCCESS != pFpcDevice->out_set(*pFpcDevice, port, OUT_B_QSFP_LPMODE, lowPwr ? 1 : 0, 0))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s Failed to set qsfpPort=%d power mode %d\n", fn, qsfpPort, lowPwr);
        return (RV_FAILURE);
    }
    if (RV_SUCCESS != pFpcDevice->out_enable(*pFpcDevice, port, OUT_B_QSFP_LPMODE, 1, 0))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s Failed to enable qsfpPort=%d output %d\n", fn, qsfpPort, OUT_B_QSFP_LPMODE);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::getQsfpCtrlLowPwr(int qsfpPort, bool& lowPwr)
{
    const char *fn = "CFpc40x16p200::getQsfpCtrlLowPwr:";
    ASSERT_QSFP_PHY_PORT(qsfpPort);
    lowPwr = false;

    uint8_t data = 0;
    uint8_t port = 0;
    convertQsfpPort(qsfpPort, port);

    reserveSpiBus();
    if (RV_SUCCESS != pFpcDevice->out_get(*pFpcDevice, port, OUT_B_QSFP_LPMODE, &data))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s failed to get qsfpPort=%d lpmode status\n", fn, qsfpPort);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    if (data != 0)
    {
        lowPwr = true;
    }
    else
    {
        lowPwr = false;
    }

    LTRACE(TRCOM|TRCLOW, "%s qsfpPort %d, lowPwr %d\n", fn, qsfpPort, lowPwr);

    return (RV_SUCCESS);
}

// When the ModSelL is held low by the host, the module responds to 2-wire
// serial communication commands. The ModSelL allows the use of multiple modules on a single
// 2-wire interface bus. When the ModSelL is "High", the module shall not respond to or
// acknowledge any 2-wire interface communication from the host.
RetVal CFpc40x16p200::setQsfpCtrlModSelI2c(int qsfpPort, bool modSelI2c)
{
    const char *fn = "CFpc40x16p200::setQsfpCtrlModSelI2c:";
    ASSERT_QSFP_PHY_PORT(qsfpPort);
    LTRACE(TRCOM|TRCLOW, "%s qsfpPort %d, modSelI2c %d\n", fn, qsfpPort, modSelI2c);

    uint8_t port = 0;
    convertQsfpPort(qsfpPort, port);

    reserveSpiBus();
    // The output values must be configured before the outputs are enabled.
    if (RV_SUCCESS != pFpcDevice->out_set(*pFpcDevice, port, OUT_A_QSFP_MODSELN, modSelI2c ? 0 : 1, 0))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s Failed to set qsfpPort=%d modSelI2c %d\n", fn, qsfpPort, modSelI2c);
        return (RV_FAILURE);
    }
    if (RV_SUCCESS != pFpcDevice->out_enable(*pFpcDevice, port, OUT_A_QSFP_MODSELN, 1, 0))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s Failed to enable qsfpPort=%d output %d\n", fn, qsfpPort, OUT_A_QSFP_MODSELN);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    return (RV_SUCCESS);
}

// A low level on the ResetL pin for longer than the minimum pulse length (t_Reset_init) 
// initiates a complete module reset, returning all user module settings to their default state.
RetVal CFpc40x16p200::setQsfpRstCtrl(int qsfpPort, bool reset)
{
    const char *fn = "CFpc40x16p200::setQsfpRstCtrl:";
    ASSERT_QSFP_PHY_PORT(qsfpPort);
    LTRACE(TRCOM|TRCLOW, "%s qsfpPort %d, reset %d\n", fn, qsfpPort, reset);

    uint8_t instNum = 0;
    uint8_t gpioPort = 0;
    if (qsfpPort == 15)           // C1
    {
        instNum = FPC_INSTANCE_3;
        gpioPort = FPC_GPIO_0;
    }
    else if (qsfpPort == 19)      // C2
    {
        instNum = FPC_INSTANCE_4;
        gpioPort = FPC_GPIO_0;
    }
    else if (qsfpPort == 1)      // L1
    {
        instNum = FPC_INSTANCE_5;
        gpioPort = FPC_GPIO_0;
    }
    else if (qsfpPort == 2)      // L2
    {
        instNum = FPC_INSTANCE_5;
        gpioPort = FPC_GPIO_1;
    }

    reserveSpiBus();
    // instance:  instance number
    // gpio_num:  gpio number, 0 to 3
    // mode:      0: Input; 1:Drive 0; 2: Drive 1
    if (RV_SUCCESS != pFpcDevice->gpio_set(*pFpcDevice, instNum, gpioPort, reset ? 2 : 1, 0))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s Failed to set qsfpPort=%d reset %d\n", fn, qsfpPort, reset);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::getQsfpRstCtrl(int qsfpPort, bool& reset)
{
    const char *fn = "CFpc40x16p200::getQsfpRstCtrl:";
    ASSERT_QSFP_PHY_PORT(qsfpPort);
    uint8_t data = 0;

    uint8_t instNum = 0;
    uint8_t gpioPort = 0;
    if (qsfpPort == 15) // C1
    {
        instNum = FPC_INSTANCE_3;
        gpioPort = FPC_GPIO_0;
    }
    else if (qsfpPort == 19)      // C2
    {
        instNum = FPC_INSTANCE_4;
        gpioPort = FPC_GPIO_0;
    }
    else if (qsfpPort == 1)      // L1
    {
        instNum = FPC_INSTANCE_5;
        gpioPort = FPC_GPIO_0;
    }
    else if (qsfpPort == 2)      // L2
    {
        instNum = FPC_INSTANCE_5;
        gpioPort = FPC_GPIO_1;
    }

    reserveSpiBus();
    if (RV_SUCCESS != pFpcDevice->gpio_get(*pFpcDevice, instNum, gpioPort, &data))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s Failed to set qsfpPort=%d\n", fn, qsfpPort);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    reset = (bool)data;

    LTRACE(TRCOM|TRCLOW, "%s qsfpPort %d, reset %d\n", fn, qsfpPort, reset);

    return (RV_SUCCESS);
}

// The FPC401 uses two sets of outputs, OUT_C[3:0] and OUT_D[3:0],
// to drive LEDs associated with the ports under its control.
// LEDs must be connected to the FPC401 in an active-low fashion.
// When the OUT_C or OUT_D pin drives a high voltage, the LED is off.
// Led state: • ON • OFF • PWM • BLINK (with programmable blink duty cycle, frequency, and ON intensity)
RetVal CFpc40x16p200::setQsfpLedState(int qsfpPort, LED_STATE state, LED_COLOR color)
{
    const char *fn = "CFpc40x16p200::setQsfpLedState:";
    ASSERT_QSFP_PHY_PORT(qsfpPort);
    LTRACE(TRCOM|TRCLOW, "%s port %d state %d color %d\n", fn, qsfpPort, state, color);

    uint8_t led_state = 0;
    uint8_t port = 0;
    convertQsfpPort(qsfpPort, port, true);

    if (state != LED_OFF)
    {
        led_state = FPC_LED_ON;
        if (state == LED_BLINK)
        {
            led_state = FPC_LED_BLINKING;
        }
    }
    else
    {
        led_state = FPC_LED_OFF;
    }

    if (led_state == FPC_LED_OFF)
    {
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_RED_LED, FPC_LED_OFF, 1, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d red LED state to OFF.\n", fn, qsfpPort);
            return (RV_FAILURE);
        }
        if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_GREEN_LED, FPC_LED_OFF, 1, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d green LED state to OFF.\n", fn, qsfpPort);
            return (RV_FAILURE);
        }
        unreserveSpiBus();
    }
    else
    {
        if (color == RED_LED)
        {
            reserveSpiBus();
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_RED_LED, led_state, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d RED LED to state %d.\n", fn, qsfpPort, led_state);
                return (RV_FAILURE);
            }
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_GREEN_LED, FPC_LED_OFF, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d GREEN LED to OFF.\n", fn, qsfpPort);
                return (RV_FAILURE);
            }
            unreserveSpiBus();
        }
        else if (color == GREEN_LED)
        {
            reserveSpiBus();
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_GREEN_LED, led_state, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d GREEN LED to state %d.\n", fn, qsfpPort, led_state);
                return (RV_FAILURE);
            }
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_RED_LED, FPC_LED_OFF, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d RED LED to OFF.\n", fn, qsfpPort);
                return (RV_FAILURE);
            }
            unreserveSpiBus();
        }
        else if (color == AMBER_LED)
        {
            reserveSpiBus();

            // set both GREEN and RED LED OFF firstly
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_GREEN_LED, FPC_LED_OFF, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d GREEN LED to state %d.\n", fn, qsfpPort, FPC_LED_OFF);
                return (RV_FAILURE);
            }
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_RED_LED, FPC_LED_OFF, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d RED LED to state %d.\n", fn, qsfpPort, FPC_LED_OFF);
                return (RV_FAILURE);
            }
            // New set next
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_GREEN_LED, led_state, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d GREEN LED to state %d.\n", fn, qsfpPort, led_state);
                return (RV_FAILURE);
            }
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, port, FPC401_SFP_RED_LED, led_state, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure QSFP port %d RED LED to state %d.\n", fn, qsfpPort, led_state);
                return (RV_FAILURE);
            }
            unreserveSpiBus();
        }
        else
        {
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s port %d not defined LED color :%d\n", fn, qsfpPort, color);
            return (RV_FAILURE);
        }
    }

    return (RV_SUCCESS);
}

// Mod_ABS is a module output pin to the Host.
// it is asserted high when the module is physically absent and deasserted low when the module is inserted.
RetVal CFpc40x16p200::getSfpStatPrest(int sfpPort, bool& sfpPrest)
{
    const char *fn = "CFpc40x16p200::getSfpStatPrest:";
    ASSERT_SFP_PORT(sfpPort);
    sfpPrest = false;

    uint8_t data = 0;

    reserveSpiBus();
    if (RV_SUCCESS != pFpcDevice->in_get(*pFpcDevice, (uint8_t)(sfpPort - 3), IN_B_MOD_ABS, &data))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: failed to get sfpPort=%d LOS status\n", fn, sfpPort);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    if (data != 0)
    {
        sfpPrest = false;
    }
    else
    {
        sfpPrest = true;
    }

    LTRACE(TRCOM|TRCLOW, "%s port %d sfpPrest %d\n", fn, sfpPort, sfpPrest);

    return (RV_SUCCESS);
}

// Rx_LOS is a module output pin to the Host.
// It indicates when high that the module optical RX input is receiving lower input power than an expected value.
RetVal CFpc40x16p200::getSfpStatLos(int sfpPort, bool& sfp_los_status)
{
    const char *fn = "CFpc40x16p200::getSfpStatLos:";
    ASSERT_SFP_PORT(sfpPort);
    sfp_los_status = false;

    uint8_t data = 0;

    reserveSpiBus();
    if (RV_SUCCESS != pFpcDevice->in_get(*pFpcDevice, (uint8_t)(sfpPort - 3), IN_C_LOS, &data))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: failed to get sfpPort=%d LOS status\n", fn, sfpPort);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    if (data != 0)
    {
        sfp_los_status = true;
    }
    else
    {
        sfp_los_status = false;
    }

    LTRACE(TRCOM|TRCLOW, "%s port %d sfp_los_status %d\n", fn, sfpPort, sfp_los_status);

    return (RV_SUCCESS);
}

// Tx_Fault is a module output pin to the Host. It indicates when high that some fault condition has been detected.
RetVal CFpc40x16p200::getSfpStatTxFault(int sfpPort, bool& sfp_txf_status)
{
    const char *fn = "CFpc40x16p200::getSfpStatTxFault:";
    ASSERT_SFP_PORT(sfpPort);
    sfp_txf_status = false;

    uint8_t data = 0;

    reserveSpiBus();
    if (RV_SUCCESS != pFpcDevice->in_get(*pFpcDevice, (uint8_t)(sfpPort - 3), IN_A_TX_FAULT, &data))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s: failed to get sfpPort=%d tx fault status\n", fn, sfpPort);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    if (data != 0)
    {
        sfp_txf_status = true;
    }
    else
    {
        sfp_txf_status = false;
    }

    LTRACE(TRCOM|TRCLOW, "%s port %d sfp_txf_status %d\n", fn, sfpPort, sfp_txf_status);

    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::setSfpRS(int sfpPort, const bool highSpeed)
{
    const char *fn = "CFpc40x16p200::setSfpRS:";
    ASSERT_SFP_PORT(sfpPort);
    LTRACE(TRCOM|TRCLOW, "%s port %d highSpeed %d\n", fn, sfpPort, highSpeed);

    for (uint8_t output_num = 0; output_num < 2; output_num++)
    {
        reserveSpiBus();
        // by default both OUT_A and OUT_B pins are disabled
        // OUT_A has default value of 1 and OUT_B has a default value of 0.
        // The output values must be configured before the outputs are enabled
        if (RV_SUCCESS != pFpcDevice->out_set(*pFpcDevice, (uint8_t)(sfpPort - 3), output_num, highSpeed ? 1 : 0, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: Failed to set sfpPort=%d rate select %d\n", fn, sfpPort, output_num);
            return (RV_FAILURE);
        }
        if (RV_SUCCESS != pFpcDevice->out_enable(*pFpcDevice, (uint8_t)(sfpPort - 3), output_num, 1, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s: Failed to enable sfpPort=%d output %d\n", fn, sfpPort, output_num);
            return (RV_FAILURE);
        }
        unreserveSpiBus();
    }

    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::getSfpRS(int sfpPort, bool& tx_sr, bool& rx_sr)
{
    uint8_t data = 0;
    const char *fn = "CFpc40x16p200::getSfpRS:";

    ASSERT_SFP_PORT(sfpPort);

    reserveSpiBus();
    if (RV_SUCCESS != pFpcDevice->out_get(*pFpcDevice, (uint8_t)(sfpPort - 3), 0, &data))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s Failed to read sfpPort=%d RS0 failed\n", fn, sfpPort);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    rx_sr = (bool)data;

    data = 0;
    reserveSpiBus();
    if (RV_SUCCESS != pFpcDevice->out_get(*pFpcDevice, (uint8_t)(sfpPort - 3), 1, &data))
    {
        unreserveSpiBus();
        LTRACE(TRCOM_ERROR|TRCHIGH, "%s Failed to read sfpPort=%d RS1 failed\n", fn, sfpPort);
        return (RV_FAILURE);
    }
    unreserveSpiBus();

    tx_sr = (bool)data;

    LTRACE(TRCOM|TRCLOW, "%s port %d tx_sr %d rx_sr %d\n", fn, sfpPort, tx_sr, rx_sr);

    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::setSfpLedState(int sfpPort, LED_STATE state, LED_COLOR color)
{
    const char *fn = "CFpc40x16p200::setSfpLedState:";
    LTRACE(TRCOM|TRCLOW, "%s port %d state %d color %d\n", fn, sfpPort, state, color);

    ASSERT_SFP_PORT(sfpPort);

    uint8_t led_state = 0;

    if (state != LED_OFF)
    {
        led_state = FPC_LED_ON;
        if (state == LED_BLINK)
        {
            led_state = FPC_LED_BLINKING;
        }
    }
    else
    {
        led_state = FPC_LED_OFF;
    }

    if (led_state == FPC_LED_OFF)
    {
        reserveSpiBus();
        if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_RED_LED, FPC_LED_OFF, 1, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d red LED state to OFF.\n", fn, sfpPort);
            return (RV_FAILURE);
        }
        if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_GREEN_LED, FPC_LED_OFF, 1, 0))
        {
            unreserveSpiBus();
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d green LED state to OFF.\n", fn, sfpPort);
            return (RV_FAILURE);
        }
        unreserveSpiBus();
    }
    else
    {
        if (color == RED_LED)
        {
            reserveSpiBus();
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_RED_LED, led_state, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d RED LED to state %d.\n", fn, sfpPort, led_state);
                return (RV_FAILURE);
            }
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_GREEN_LED, FPC_LED_OFF, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d GREEN LED to OFF.\n", fn, sfpPort);
                return (RV_FAILURE);
            }
            unreserveSpiBus();
        }
        else if (color == GREEN_LED)
        {
            reserveSpiBus();
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_GREEN_LED, led_state, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d GREEN LED to state %d.\n", fn, sfpPort, led_state);
                return (RV_FAILURE);
            }
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_RED_LED, FPC_LED_OFF, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d RED LED to OFF.\n", fn, sfpPort);
                return (RV_FAILURE);
            }
            unreserveSpiBus();
        }
        else if (color == AMBER_LED)
        {
            reserveSpiBus();
            // set both GREEN and RED LED OFF firstly
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_GREEN_LED, FPC_LED_OFF, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d GREEN LED to state %d.\n", fn, sfpPort, FPC_LED_OFF);
                return (RV_FAILURE);
            }
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_RED_LED, FPC_LED_OFF, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d RED LED to state %d.\n", fn, sfpPort, FPC_LED_OFF);
                return (RV_FAILURE);
            }
            // new set next
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_GREEN_LED, led_state, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d GREEN LED to state %d.\n", fn, sfpPort, led_state);
                return (RV_FAILURE);
            }
            if (RV_SUCCESS != pFpcDevice->led_set(*pFpcDevice, (uint8_t)(sfpPort - 3), FPC401_SFP_RED_LED, led_state, 1, 0))
            {
                unreserveSpiBus();
                LTRACE(TRCOM_ERROR|TRCHIGH, "%s fail to configure SFP port %d RED LED to state %d.\n", fn, sfpPort, led_state);
                return (RV_FAILURE);
            }
            unreserveSpiBus();
        }
        else
        {
            LTRACE(TRCOM_ERROR|TRCHIGH, "%s port %d not defined LED color :%d\n", fn, sfpPort, color);
            return (RV_FAILURE);
        }
    }

    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::downstream_read(uint8_t port, uint8_t device, uint8_t offset, uint8_t *data)
{
    if (pFpcDevice->downstream_read(*pFpcDevice, port, device, offset, data) != EOK)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "CFpc40x16p200::downstream_read: fail to read port %d register %x\n", port, offset);
        return (RV_FAILURE);
    }
    LTRACE(TRCOM|TRCLOW, "CFpc40x16p200::downstream_read: port %d device %d offset %x data %x\n", port, device, offset, *data);
    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::downstream_write(uint8_t port, uint8_t device, uint8_t offset, uint8_t data)
{
    LTRACE(TRCOM|TRCLOW, "CFpc40x16p200::downstream_write: port %d device %d offset %x data %x\n", port, device, offset, data);
    if (pFpcDevice->downstream_write(*pFpcDevice, port, device, offset, data) != EOK)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "CFpc40x16p200::downstream_write: fail to write port %d register %x with data %x\n",
               port, offset, data);
        return (RV_FAILURE);
    }
    return (RV_SUCCESS);
}

RetVal CFpc40x16p200::dumpRegisters()
{
    uint8_t data = 0;

    for (uint16_t dev = FPC_INSTANCE_0; dev < FPC_INSTANCE_MAX; dev++)
    {
        LTRACE(TRCOM_ERROR|TRCHIGH, "dump FPC401 %d registers\n", dev);
        for (uint16_t reg = FPC401_RESET; reg <= FPC401_DEVICE_ID; reg++)
        {
            reserveSpiBus();
            pFpcDevice->read_fn(dev, reg, &data);
            unreserveSpiBus();
            if (reg % 2 == 0)
            {
                LTRACE(TRCOM_ERROR|TRCHIGH, "dev %d regAddr: 0x%x regData: 0x%x       ", dev, reg, data);
            }
            else
            {
                LTRACE(TRCOM_ERROR|TRCHIGH, "dev %d regAddr: 0x%x regData: 0x%x\n", dev, reg, data);
            }
        }
    }

    return (RV_SUCCESS);
}

}
